package org.ids_mannheim;

import org.apache.commons.io.IOUtils;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.io.*;
import java.util.ArrayList;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Logger;

import static org.junit.jupiter.api.Assertions.*;

class WorkerTest {
    private ByteArrayOutputStream errContent;
    private final PrintStream originalErr = System.err;
    Worker worker;
    ConcurrentHashMap<String, AtomicInteger> map;
    @BeforeEach
    public void setUpStreams() {
        errContent = new ByteArrayOutputStream();
        System.setErr(new PrintStream(errContent));
    }

    @AfterEach
    public void restoreStreams() {
        System.setErr(originalErr);
    }

    @Test
    void resultAndOutputAreCorrectWithGz() throws IOException {
        Map<String, Integer> gold = Map.of(
                "und	und	KON	Fluchen	Fluchen	NN", 1,
                "Bestreben	Bestreben	NN	,	,	$,", 1,
                "Bürger	Bürger	NN	sich	sich	PRF", 1,
                "dieses	dies	PDAT	würdigen	würdig	ADJA", 1,
                "im	in	APPRART	Kriegshandwerk	Kriegshandwerk	NN", 1,
                "man	man	PIS	nur	nur	ADV", 1,
                "von	von	APPR	Longwy	--	NE", 3,
                "nicht	nicht	PTKNEG	ungeschickt	ungeschickt	ADJD", 1,
                "Republikaner	Republikaner	NN	und	und	KON", 1,
                "Patriotismus	Patriotismus	NN	derer	die	PDS", 1
        );

        File tempFile = File.createTempFile("goe_sample", ".conllu.gz");
        tempFile.deleteOnExit();
        try (FileOutputStream out = new FileOutputStream(tempFile)) {
            IOUtils.copy(Thread.currentThread().getContextClassLoader()
                    .getResourceAsStream("goe_sample.conllu.gz"), out);
        }
        ArrayList<String> fnames = new ArrayList<>();
        fnames.add(tempFile.getAbsolutePath());
        map = new ConcurrentHashMap<>();
        LinkedBlockingQueue<Integer> queue = new LinkedBlockingQueue<>(2);
        worker = new Worker(
                queue,
                fnames,
                2,
                8,
                10,
                map,
                true,
                false,
                new WorkerNodePool(""),
                new Progressbar(tempFile.length()),
                Logger.getLogger(TotalNGrams.class.getSimpleName()));

        queue.add(0);
        queue.add(-1);
        worker.run();
        gold.forEach((key, value) -> {
            AtomicInteger observed = map.get(key);
            assertNotNull(observed);
            if (observed != null) {
                assertEquals(value, observed.intValue());
            }
        });
        assertTrue(errContent.toString().contains("with 1 text"));
        assertTrue(errContent.toString().contains("100%"));
    }

    @Test
    void downcasedResultAndOutputAreCorrect() throws IOException {
        Map<String, Integer> gold = Map.of(
                "und	und	KON	fluchen	Fluchen	NN", 1,
                "bestreben	Bestreben	NN	,	,	$,", 1,
                "bürger	Bürger	NN	sich	sich	PRF", 1,
                "dieses	dies	PDAT	würdigen	würdig	ADJA", 1,
                "im	in	APPRART	kriegshandwerk	Kriegshandwerk	NN", 1,
                "man	man	PIS	nur	nur	ADV", 1,
                "von	von	APPR	longwy	--	NE", 3,
                "nicht	nicht	PTKNEG	ungeschickt	ungeschickt	ADJD", 1,
                "republikaner	Republikaner	NN	und	und	KON", 1,
                "patriotismus	Patriotismus	NN	derer	die	PDS", 1
        );

        File tempFile = File.createTempFile("goe_sample", ".conllu.gz");
        tempFile.deleteOnExit();
        try (FileOutputStream out = new FileOutputStream(tempFile)) {
            IOUtils.copy(Thread.currentThread().getContextClassLoader()
                    .getResourceAsStream("goe_sample.conllu.gz"), out);
        }
        ArrayList<String> fnames = new ArrayList<>();
        fnames.add(tempFile.getAbsolutePath());
        map = new ConcurrentHashMap<>();
        LinkedBlockingQueue<Integer> queue = new LinkedBlockingQueue<>(2);
        worker = new Worker(
                queue,
                fnames,
                2,
                8,
                10,
                map,
                true,
                true,
                new WorkerNodePool(""),
                new Progressbar(tempFile.length()),
                Logger.getLogger(TotalNGrams.class.getSimpleName()));

        queue.add(0);
        queue.add(-1);
        worker.run();
        gold.forEach((key, value) -> {
//            System.out.println("KEY: " + key);
            AtomicInteger observed = map.get(key);
            assertNotNull(observed);
            if (observed != null) {
                assertEquals(value, observed.intValue());
            }
        });
        assertTrue(errContent.toString().contains("100%"));
    }

    @Test
    void resultAndOutputAreCorrectForEntities() throws IOException {
        Map<String, Integer> gold = Map.of(
                "\"\t\"\t$(", 1,
                "\'\t\"\t$(", 1,
                "&\t&amp;\tKON", 1,
                "<\t&lt;\t$(", 1,
                ">\t&gt;\t$(", 1
        );

        File tempFile = File.createTempFile("entities", ".conllu");
        tempFile.deleteOnExit();
        try (FileOutputStream out = new FileOutputStream(tempFile)) {
            IOUtils.copy(Thread.currentThread().getContextClassLoader()
                    .getResourceAsStream("entities.conllu"), out);
        }
        ArrayList<String> fnames = new ArrayList<>();
        fnames.add(tempFile.getAbsolutePath());
        map = new ConcurrentHashMap<>();
        LinkedBlockingQueue<Integer> queue = new LinkedBlockingQueue<>(2);
        worker = new Worker(
                queue,
                fnames,
                1,
                5,
                10,
                map,
                true,
                false,
                new WorkerNodePool(""),
                new Progressbar(tempFile.length()),
                Logger.getLogger(TotalNGrams.class.getSimpleName()));

        queue.add(0);
        queue.add(-1);
        worker.run();
        gold.forEach((key, value) -> assertNotNull(map.get(key)));
        gold.forEach((key, value) -> assertEquals(value, map.get(key).intValue()));
    }

    @Test
    void resultAndOutputAreCorrectWithXz() throws IOException {
        Map<String, Integer> gold = Map.of(
                "und	und	KON	Fluchen	Fluchen	NN", 1,
                "Bestreben	Bestreben	NN	,	,	$,", 1,
                "Bürger	Bürger	NN	sich	sich	PRF", 1,
                "dieses	dies	PDAT	würdigen	würdig	ADJA", 1,
                "im	in	APPRART	Kriegshandwerk	Kriegshandwerk	NN", 1,
                "man	man	PIS	nur	nur	ADV", 1,
                "von	von	APPR	Longwy	--	NE", 3,
                "nicht	nicht	PTKNEG	ungeschickt	ungeschickt	ADJD", 1,
                "Republikaner	Republikaner	NN	und	und	KON", 1,
                "Patriotismus	Patriotismus	NN	derer	die	PDS", 1
        );

        File tempFile = File.createTempFile("goe_sample", ".conllu.xz");
        tempFile.deleteOnExit();
        try (FileOutputStream out = new FileOutputStream(tempFile)) {
            IOUtils.copy(Thread.currentThread().getContextClassLoader()
                    .getResourceAsStream("goe_sample.conllu.xz"), out);
        }
        ArrayList<String> fnames = new ArrayList<>();
        fnames.add(tempFile.getAbsolutePath());
        map = new ConcurrentHashMap<>();
        LinkedBlockingQueue<Integer> queue = new LinkedBlockingQueue<>(2);
        worker = new Worker(
                queue,
                fnames,
                2,
                8,
                10,
                map,
                true,
                false,
                new WorkerNodePool(""),
                new Progressbar(tempFile.length()),
                Logger.getLogger(TotalNGrams.class.getSimpleName()));

        queue.add(0);
        queue.add(-1);
        worker.run();
        gold.forEach((key, value) -> assertEquals(value, map.get(key).intValue()));
    }

}
